from rest_framework import serializers

from .mixins import BridgerSerializerFieldMixin
from .types import BridgerType


class IntegerField(BridgerSerializerFieldMixin, serializers.IntegerField):
    field_type = BridgerType.NUMBER.value

    def get_representation(self, request, field_name):
        representation = super().get_representation(request, field_name)
        representation["precision"] = 0
        return representation


class DecimalField(BridgerSerializerFieldMixin, serializers.DecimalField):
    field_type = BridgerType.NUMBER.value

    def __init__(self, *args, **kwargs):
        self.percent = kwargs.pop("percent", False)
        super().__init__(*args, **kwargs)

    # TODO: If this is used, then the validation for max_digits and decimal_fields is not done
    # def validate_precision(self, value):
    #     return value

    def get_representation(self, request, field_name):
        representation = super().get_representation(request, field_name)
        representation["precision"] = self.decimal_places

        if self.percent:  # TODO: Discuss with Christoph if this is necessary like this
            representation["type"] = BridgerType.PERCENT.value
            representation["precision"] = self.decimal_places - 2

        return representation


class FloatField(BridgerSerializerFieldMixin, serializers.FloatField):
    field_type = BridgerType.NUMBER.value

    def __init__(self, *args, **kwargs):
        self.decimal_places = kwargs.pop("decimal_places", 2)
        super().__init__(*args, **kwargs)

    def get_representation(self, request, field_name):
        representation = super().get_representation(request, field_name)
        representation["precision"] = self.decimal_places
        return representation



# Generated by Django 2.2.11 on 2020-03-27 10:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("bridger", "0003_auto_20200218_1329"),
    ]

    operations = [
        migrations.AddField(model_name="notification", name="endpoint", field=models.URLField(null=True),),
        migrations.AlterField(
            model_name="notification",
            name="timestamp_created",
            field=models.DateTimeField(auto_now_add=True, verbose_name="Created"),
        ),
        migrations.AlterField(
            model_name="notification",
            name="timestamp_mailed",
            field=models.DateTimeField(blank=True, null=True, verbose_name="Mailed"),
        ),
        migrations.AlterField(
            model_name="notification",
            name="timestamp_read",
            field=models.DateTimeField(blank=True, null=True, verbose_name="Read"),
        ),
        migrations.AlterField(
            model_name="notification",
            name="timestamp_received",
            field=models.DateTimeField(blank=True, null=True, verbose_name="Received"),
        ),
    ]



from typing import Optional

from rest_framework.request import Request
from bridger.metadata.mixins import BridgerViewSetConfig


class TitleConfig(BridgerViewSetConfig):
    def get_instance_title(self) -> str:
        name = self.view.get_model()._meta.verbose_name

        if self.instance:
            return f"{name}: {str(self.view.get_object())}"
        return name

    def get_delete_title(self) -> str:
        name = self.view.get_model()._meta.verbose_name

        if self.instance:
            return f"Delete {name}: {str(self.view.get_object())}"
        return f"Delete {name}"

    def get_list_title(self) -> str:
        return self.view.get_model()._meta.verbose_name_plural

    def get_create_title(self) -> str:
        return f"Create {self.view.get_model()._meta.verbose_name}"

    def get_metadata(self):
        yield "instance", self.get_instance_title()
        yield "list", self.get_list_title()
        yield "create", self.get_create_title()
        yield "delete", self.get_delete_title()

    @classmethod
    def get_metadata_key(cls):
        return "titles"

    @classmethod
    def get_metadata_fieldname(cls):
        return "title_config_class"


class TitleConfigMixin:
    title_config_class = TitleConfig



from rest_framework.reverse import reverse

from bridger.buttons import ActionButton, ButtonConfig
from bridger.display import FieldSet, InstanceDisplay, Section
from bridger.enums import WBIcon
from bridger.settings import bridger_settings


def share_action_button(request):
    serializer = bridger_settings.DEFAULT_SHARE_SERIALIZER
    fields = tuple([field for field in serializer().fields])

    btn = ActionButton(
        icon=WBIcon.TRADE.value,
        description_fields="",
        endpoint=reverse("bridger:share", request=request),
        instance_display=InstanceDisplay(sections=tuple([Section(fields=FieldSet(fields=fields))])),
        confirm_config=ButtonConfig(label="Share", icon=WBIcon.TRADE.value),
        serializer=serializer,
    )
    btn.request = request
    return btn



from rest_framework.routers import DefaultRouter, DynamicRoute, Route


class BridgerRouter(DefaultRouter):
    """
    The default Django Rest Framework router with a destroy_multiple method
    """

    routes = [
        # List route.
        Route(
            url=r"^{prefix}{trailing_slash}$",
            mapping={"get": "list", "post": "create", "delete": "destroy_multiple"},
            name="{basename}-list",
            detail=False,
            initkwargs={"suffix": "List"},
        ),
        # Dynamically generated list routes. Generated using
        # @action(detail=False) decorator on methods of the viewset.
        DynamicRoute(url=r"^{prefix}/{url_path}{trailing_slash}$", name="{basename}-{url_name}", detail=False, initkwargs={},),
        # Detail route.
        Route(
            url=r"^{prefix}/{lookup}{trailing_slash}$",
            mapping={"get": "retrieve", "put": "update", "patch": "partial_update", "delete": "destroy",},
            name="{basename}-detail",
            detail=True,
            initkwargs={"suffix": "Instance"},
        ),
        # Dynamically generated detail routes. Generated using
        # @action(detail=True) decorator on methods of the viewset.
        DynamicRoute(
            url=r"^{prefix}/{lookup}/{url_path}{trailing_slash}$", name="{basename}-{url_name}", detail=True, initkwargs={},
        ),
        Route(
            url=r"^{prefix}/{lookup}/history{trailing_slash}$",
            mapping={"get": "history_list"},
            name="{basename}-history-list",
            detail=False,
            initkwargs={"suffix": "History List", "historical_mode": True},
        ),
        Route(
            url=r"^{prefix}/{lookup}/history/(?P<history_id>[^/.]+){trailing_slash}$",
            mapping={"get": "history_retrieve",},
            name="{basename}-history-detail",
            detail=True,
            initkwargs={"suffix": "History Instance", "historical_mode": False},
        ),
    ]



# Generated by Django 3.0.6 on 2020-05-14 15:11

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("bridger", "0007_asset"),
    ]

    operations = [
        migrations.RenameModel(old_name="Asset", new_name="Asset111",),
    ]



import functools
import logging
from inspect import getmembers

from rest_framework import serializers
from rest_framework.reverse import reverse

from bridger.signals.instance_buttons import add_additional_resource

from .mixins import BridgerSerializerFieldMixin
from .types import BridgerType

logger = logging.getLogger(__name__)


def register_resource():
    def decorator(func):
        func._is_additional_resource = True
        return func

    return decorator


def _is_additional_resource(attr):
    return hasattr(attr, "_is_additional_resource")


def register_dynamic_button():
    def decorator(func):
        func._is_dynamic_button = True
        return func

    return decorator


def _is_dynamic_button(attr):
    return hasattr(attr, "_is_dynamic_button")


class DynamicButtonField(BridgerSerializerFieldMixin, serializers.ReadOnlyField):

    field_type = "_dynamic_buttons"

    def get_attribute(self, instance):
        return instance

    def to_representation(self, value):
        request = self.parent.context["request"]
        buttons = list()

        for _, button_func in getmembers(self.parent.__class__, _is_dynamic_button):
            btns = button_func(self.parent, instance=value, request=request, user=request.user)
            buttons.extend([dict(btn) for btn in btns])

        return buttons


class AdditionalResourcesField(BridgerSerializerFieldMixin, serializers.ReadOnlyField):

    field_type = "_additional_resources"

    def get_attribute(self, instance):
        return instance

    def to_representation(self, value):
        request = self.parent.context["request"]
        resources = dict()
        for _, function in getmembers(self.parent.__class__, _is_additional_resource):
            _d = function(self.parent, instance=value, request=request, user=request.user)
            resources = {**resources, **_d}

        remote_resources = add_additional_resource.send(
            sender=self.parent.__class__, serializer=self.parent, instance=value, request=request, user=request.user
        )
        for _, resource in remote_resources:
            resources = {**resources, **resource}

        return resources


class HyperlinkField(BridgerSerializerFieldMixin, serializers.ReadOnlyField):
    def __init__(self, *args, **kwargs):
        self.reverse_name = kwargs.pop("reverse_name")
        super().__init__(*args, **kwargs)

    def get_attribute(self, obj):
        request = self.context.get("request", None)
        if request:
            return reverse(self.reverse_name, args=[obj.id], request=request)
        return reverse(self.reverse_name, args=[obj.id])


class ReadOnlyField(BridgerSerializerFieldMixin, serializers.ReadOnlyField):
    field_type = BridgerType.TEXT.value


class SerializerMethodField(BridgerSerializerFieldMixin, serializers.SerializerMethodField):
    def __init__(self, method_name=None, field_type=BridgerType.TEXT.value, **kwargs):
        self.field_type = field_type
        super().__init__(method_name, **kwargs)



from django.db.models import F, Value
from django.db.models.functions import Concat
from django.utils import timezone
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.reverse import reverse

from bridger import buttons as bt
from bridger import display as dp
from bridger import viewsets
from bridger.enums import Button, RequestType

from .serializers import get_historical_serializer


def get_historical_viewset(model, historical_model):
    class HistoricalModelViewSet(viewsets.ModelViewSet):

        BUTTONS = [Button.REFRESH.value]
        CREATE_BUTTONS = []

        def get_custom_instance_buttons(self, request, buttons):
            return []

        def get_custom_list_instance_buttons(self, request, buttons):
            return []

        LIST_DISPLAY = dp.ListDisplay(
            fields=[
                dp.Field(key="history_date", label="Changed"),
                dp.Field(key="history_change_reason", label="Reason"),
                dp.Field(key="history_type", label="Type"),
                dp.Field(key="user_repr", label="User"),
            ]
        )

        def get_instance_endpoint(self, request, endpoint=None):
            return (
                f"{model.get_endpoint_basename()}-history-list",
                [self.kwargs["model_pk"]],
                {},
            )

        queryset = historical_model.objects.all()
        serializer_class = get_historical_serializer(historical_model)

        ordering_fields = ("history_id",)

        def get_queryset(self):
            qs = super().get_queryset()
            qs = qs.annotate(user_repr=Concat(F("history_user__first_name"), Value(" "), F("history_user__last_name"),))
            return qs

    return HistoricalModelViewSet



from rest_framework import serializers

from .mixins import BridgerSerializerFieldMixin
from .types import BridgerType


class BooleanField(BridgerSerializerFieldMixin, serializers.BooleanField):
    field_type = BridgerType.BOOLEAN.value



from typing import Tuple

from rest_framework.request import Request
from rest_framework.views import View

class BridgerViewSetConfig:

    def __init__(self, view: View, request: Request, instance: bool):
        self.view = view
        self.request = request
        self.instance = instance

    @classmethod
    def get_metadata_parameters(cls) -> Tuple[str, str]:
        """ Returns the parameters needed to get the metadata from a config

        Parameters: Fieldname, Methodname
        """
        raise NotImplementedError("This needs to be implemented.")

    @classmethod
    def get_metadata_key(cls):
        raise NotImplementedError()

    @classmethod
    def get_metadata_fieldname(cls):
        raise NotImplementedError()

    @classmethod
    def get_metadata_method(cls):
        return "get_metadata"



from django.conf import settings
from django.urls.exceptions import NoReverseMatch
from rest_framework import status, viewsets, filters
from rest_framework.decorators import api_view, permission_classes
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework.pagination import CursorPagination
from django_filters.rest_framework import DjangoFilterBackend

def get_full_widget_url(reverse_path):
    return  f'{settings.BASE_ENDPOINT_URL}?widget_endpoint={settings.BASE_ENDPOINT_URL}{reverse_path}'

def set_identifier(identifier):
    def decorator(func):
        func.identifier = identifier
        return func

    return decorator

@api_view(["GET"])
@permission_classes([IsAuthenticated])
def api_endpoints_root(request, format=None):
    try:
        endpoints = dict()
        for wb_endpoint in settings.WB_ENDPOINTS:
            try:
                endpoints[wb_endpoint] = reverse(
                    f"{wb_endpoint}:api-root", request=request, format=format
                )
            except NoReverseMatch:
                pass

        return Response(endpoints)

    except AttributeError:
        return Response(
            {"error": "No Endpoints specified."}, status=status.HTTP_400_BAD_REQUEST
        )



from enum import Enum
from bridger.utils.enum import ChoiceEnum

class WBColor(Enum):
    GREEN_LIGHT = 'rgb(214, 229, 145)'
    GREEN = 'rgb(182, 223, 18)'
    GREEN_DARK = 'rgb(111, 181, 68)'
    RED_LIGHT = 'rgb(244, 132, 116)'
    RED = 'rgb(250, 84, 62)'
    RED_DARK = 'rgb(224, 57, 49)'
    YELLOW_LIGHT = 'rgb(247, 232, 113)'
    YELLOW = 'rgb(247, 207, 77)'
    YELLOW_DARK = 'rgb(247, 190, 47)'
    GREY = 'rgb(133, 144, 162)'
    BLUE_LIGHT = 'rgb(104, 198, 224)'
    BLUE = 'rgb(69, 112, 224)'
    BLUE_DARK = 'rgb(33, 26, 233)'

class ColorProduct(ChoiceEnum):
    BLACK_1="#3C4859"
    BLACK_2="#6D7683"
    BLACK_3="#9EA3AC"
    BLACK_4="#CED1D5"
    DARK_BLUE_1="#0585FF"
    DARK_BLUE_2="#69B6FF"
    DARK_BLUE_3="#B4DAFF"
    LIGHT_BLUE_1="#70D6FF"
    LIGHT_BLUE_2="#A9E7FF"
    LIGHT_BLUE_3="#D4F3FF"
    TURQUOISE_DARK_1="#01ABAA"
    TURQUOISE_DARK_2="#67CDCC"
    TURQUOISE_DARK_3="#B3E6E5"
    TURQUOISE_LIGHT_1="#05D6A1"
    TURQUOISE_LIGHT_2="#69E7C6"
    TURQUOISE_LIGHT_3="#ADF1E0"
    GREEN_1="#8CD867"
    GREEN_2="#BAE8A4"
    GREEN_3="#DDF3D1"
    PURPLE_DARK_1="#5624DA"
    PURPLE_DARK_2="#9A7CE9"
    PURPLE_DARK_3="#CCBDF4"
    PURPLE_LIGHT_1="#A258E5"
    PURPLE_LIGHT_2="#C79BEF"
    PURPLE_LIGHT_3="#E3CDF7"
    PINK_1="#EF476F"
    PINK_2="#F591A9"
    PINK_3="#FAC8D4"
    YELLOW_1="#FFD166"
    YELLOW_2="#FFE8B2" 

    @classmethod
    def get_gradient(cls, color):
        colors = cls.values()
        color_value = cls[color].value
        index = colors.index(color_value)
        if index == -1:
            return colors
        else:
            return colors[index:] + colors[:index]

class TransparencyMixin:

    def get_cell_formatting(self, request):
        return [
            {
                'column': None,
                'conditions': [
                    {
                        'condition': None,
                        'style': {
                            'backgroundColor': f'rgba(255, 255, 255, {getattr(self, "TRANSPARENCY", 0.4)})'
                        }
                    },
                ]
            }
        ]



from contextlib import suppress
from typing import List, Set, Dict

from rest_framework.request import Request

from bridger.enums import Button
from bridger.metadata.mixins import BridgerViewSetConfig

class BaseButtonConfig:

    FSM_LIST = True
    FSM_INSTANCE = True
    FSM_DROPDOWN = False
    FSM_WEIGHT = 100
    FSM_BUTTONS = set()

    LIST_BUTTONS = {Button.NEW.value, Button.REFRESH.value}
    LIST_BUTTONS_ORDERING = [Button.NEW.value, Button.REFRESH.value]

    INSTANCE_BUTTONS = {Button.SAVE.value, Button.REFRESH.value, Button.DELETE.value}
    INSTANCE_BUTTONS_ORDERING = [Button.SAVE.value, Button.REFRESH.value, Button.DELETE.value]

    CREATE_BUTTONS = {Button.SAVE.value, Button.SAVE_AND_CLOSE.value, Button.SAVE_AND_NEW.value, Button.RESET.value}
    CREATE_BUTTONS_ORDERING = [Button.SAVE.value, Button.SAVE_AND_CLOSE.value, Button.SAVE_AND_NEW.value, Button.RESET.value]

    CUSTOM_INSTANCE_BUTTONS = set()
    CUSTOM_LIST_INSTANCE_BUTTONS = set()
    CUSTOM_BUTTONS = set()

    def get_custom_instance_buttons(self, request: Request) -> List:
        return []

    def get_list_instance_buttons(self, request: Request) -> List:
        return []

    def order_buttons(self, buttons: Set, ordering: List) -> List:
        yield from filter(lambda b: b in buttons, ordering)

    def get_list_buttons(self, request: Request) -> List:
        content_type = self.viewset.get_content_type()
        add_permission = f"{content_type.app_label}.add_{content_type.model}"

        if not request.user.has_perm(add_permission):
            with suppress(KeyError):
                self.LIST_BUTTONS.remove(Button.NEW.value)

        yield from self.order_buttons(self.LIST_BUTTONS, self.LIST_BUTTONS_ORDERING)

    def get_instance_buttons(self, request: Request) -> List:
        content_type = self.viewset.get_content_type()
        change_permission = f"{content_type.app_label}.change_{content_type.model}"
        delete_permission = f"{content_type.app_label}.delete_{content_type.model}"

        if not request.user.has_perm(change_permission):
            with suppress(KeyError):
                self.INSTANCE_BUTTONS.remove(Button.SAVE.value)

        if not request.user.has_perm(delete_permission):
            with suppress(KeyError):
                self.INSTANCE_BUTTONS.remove(Button.DELETE.value)

        yield from self.order_buttons(self.INSTANCE_BUTTONS, self.INSTANCE_BUTTONS_ORDERING)

    def get_create_buttons(self, request: Request) -> List:
        yield from self.order_buttons(self.CREATE_BUTTONS, self.CREATE_BUTTONS_ORDERING)

    def to_metadata(self, request: Request) -> Dict:
        yield "instance", self.get_instance_buttons(request)
        yield "list", self.get_list_buttons(request)
        yield "create", self.get_create_buttons(request)
        yield "custom_instance", self.get_custom_instance_buttons(request)

    def determine_metadata(self) -> Dict:
        return {"buttons": self.to_metadata()}


class ButtonViewSetMixin:
    button_config_class = BaseButtonConfig



from bridger import serializers
from bridger.buttons import WidgetButton


def get_historical_serializer(historical_model):
    class HistoricalModelSerializer(serializers.ModelSerializer):
        user_repr = serializers.SerializerMethodField()

        def get_user_repr(self, obj):
            user = obj.history_user
            return f"{user.first_name} {user.last_name}"

        class Meta:
            model = historical_model
            fields = read_only_fields = (
                "history_id",
                "history_date",
                "history_change_reason",
                "history_type",
                "history_user",
                "user_repr",
            )

    return HistoricalModelSerializer



from django.urls import include, path
from django.views.generic import TemplateView

from .clubhouse import ClubHouseView
from .frontend_user_configuration import FrontendUserConfigurationModelViewSet
from .markdown.views import AssetCreateView, AssetRetrieveView, BlockDiag, TemplateTagView
from .menus.views import MenuAPIView
from .notifications.viewsets.viewsets import NotificationModelViewSet
from .profile import UserViewSet
from .routers import BridgerRouter
from .share.views import ShareAPIView
from .tags.viewsets import TagRepresentationViewSet
from .views import Config, Profile

router = BridgerRouter()
router.register(
    r"frontenduserconfiguration", FrontendUserConfigurationModelViewSet, basename="frontenduserconfiguration",
)
router.register(r"notification", NotificationModelViewSet, basename="notification")
router.register(r"user", UserViewSet, basename="user")
router.register(r"clubhouse", ClubHouseView, basename="clubhouse")
router.register(r"tagrepresentation", TagRepresentationViewSet, basename="tagrepresentation")

urlpatterns = [
    path("config/", Config.as_view(), name="config"),
    path("profile/", Profile.as_view(), name="profile"),
    path("menu/", MenuAPIView.as_view(), name="menu"),
    path("share/", ShareAPIView.as_view(), name="share"),
    path("markdown/blockdiag/", BlockDiag.as_view(), name="blockdiag"),
    path("markdown/asset/", AssetCreateView.as_view(), name="markdown-asset-upload"),
    path("markdown/asset/<uuid>/", AssetRetrieveView.as_view(), name="asset-retrieve"),
    path("markdown/templatetag/", TemplateTagView.as_view(), name="markdown-tags"),
    path("", include(router.urls)),
]




import pytest
from bridger.tests.mixins import TestModelClass, TestSerializerClass, TestrepresentationViewSetClass, TestViewSetClass
from bridger.tests.utils import get_all_subclasses, is_intermediate_table_m2m, all_subclasses
from bridger.tests.signals import get_specfics_module
from bridger import serializers
from bridger import viewsets
from django.db import models
from django.urls import get_resolver
get_resolver().url_patterns

models = filter(lambda x: not x.__module__.startswith(('bridger', 'django', 'rest_framework', 'dynamic_preferences', 'eventtools')) 
                and not x._meta.abstract and not is_intermediate_table_m2m(x), get_all_subclasses(models.Model))
serializers = filter(lambda x: "bridger" not in x.__module__, get_all_subclasses(serializers.ModelSerializer))
representationviewsets = filter(lambda x: "bridger" not in x.__module__, get_all_subclasses(viewsets.RepresentationViewSet))
modelviewsets = filter(lambda x: "bridger" not in x.__module__, get_all_subclasses(viewsets.ModelViewSet))
inf_modelviewsets = filter(lambda x: "bridger" not in x.__module__, get_all_subclasses(viewsets.InfiniteDataModelView))

remote_models = get_specfics_module.send(sender = models)
if remote_models:
    _, models = remote_models[0]
remote_serializers = get_specfics_module.send(sender = serializers)
if remote_serializers:
    _, serializers = remote_serializers[0]
remote_representationviewsets = get_specfics_module.send(sender = representationviewsets)
if remote_representationviewsets:
    _, representationviewsets = remote_representationviewsets[0]
remote_modelviewsets = get_specfics_module.send(sender = modelviewsets)
if remote_modelviewsets:
    _, modelviewsets = remote_modelviewsets[0]
remote_inf_modelviewsets = get_specfics_module.send(sender = inf_modelviewsets)
if remote_inf_modelviewsets:
    _, inf_modelviewsets = remote_inf_modelviewsets[0]

if len([imvs for imvs in inf_modelviewsets]) == 0:
    inf_modelviewsets = [None]

@pytest.mark.django_db 
class TestProject:
    @pytest.mark.parametrize("model", models)
    def test_models(self, model):
        # print(mvs.__name__)
        # print(mvs.__module__)
        # print(model.__dict__)
        my_test = TestModelClass(model)
        my_test.execute_test()

    @pytest.mark.parametrize("serializer", serializers)
    def test_serializers(self, serializer):
        my_test = TestSerializerClass(serializer)
        my_test.execute_test()

    @pytest.mark.parametrize("rvs", representationviewsets)
    def test_representationviewsets(self, rvs, admin_client):
        my_test = TestrepresentationViewSetClass(rvs)
        my_test.execute_test(admin_client)

    @pytest.mark.parametrize("mvs", modelviewsets)
    def test_modelviewsets(self, mvs, admin_client):
        my_test = TestViewSetClass(mvs)
        my_test.execute_test(admin_client)

    @pytest.mark.parametrize("imvs", inf_modelviewsets)
    def test_inf_modelviewsets(self, imvs, admin_client):
        if imvs:
            my_test = TestViewSetClass(imvs)
            my_test.execute_test(admin_client)


import pathlib
import uuid

from django.db import models
from django.dispatch import receiver


class Asset(models.Model):

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    file = models.FileField(upload_to="markdown/assets")
    content_type = models.CharField(max_length=10, null=True, blank=True)
    file_url_name = models.CharField(max_length=1024, null=True, blank=True)
    # public = models.BooleanField(default=True)

    class Meta:
        verbose_name = "Asset"
        verbose_name_plural = "Assets"


@receiver(models.signals.pre_save, sender="bridger.Asset")
def generate_content_type(sender, instance, **kwargs):
    if suffix := pathlib.Path(instance.file.name).suffix:
        instance.content_type = suffix[1:]
        instance.file_url_name = f"{instance.id}{suffix}"



import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework.views import APIView

from .buttons import ActionButton
from .display import FieldSet, InstanceDisplay, Section
from .enums import AuthType
from .menus import default_registry
from .settings import bridger_settings
from .share import ShareSerializer

logger = logging.getLogger(__name__)


class Profile(APIView):
    permission_classes = [IsAuthenticated]

    def get(self, request: Request) -> Response:
        return Response(
            {
                "config": reverse("bridger:frontenduserconfiguration-list", request=request),
                "name": f"{request.user.first_name} {request.user.last_name}",
                "email": request.user.email,
                "profile": bridger_settings.PROFILE(request),
            }
        )


class Config(APIView):
    permission_classes = []

    def get(self, request: Request) -> Response:
        menu_calendar = None
        if bridger_settings.FRONTEND_MENU_CALENDAR:
            menu_calendar = bridger_settings.FRONTEND_MENU_CALENDAR(request=request)

        return Response(
            {
                "authentication": bridger_settings.DEFAULT_AUTH_CONFIG(request),
                "profile": reverse("bridger:profile", request=request),
                "menu": reverse("bridger:menu", request=request),
                "notification": bridger_settings.DEFAULT_NOTIFICATION_CONFIG(request=request),
                "share": dict(bridger_settings.DEFAULT_SHARE_BUTTON(request=request)),
                "markdown": {"blockdiag": reverse("bridger:blockdiag", request=request)},
                "clubhouse": bridger_settings.CLUBHOUSE_CONFIG(request),
                "menu_calendar": menu_calendar,
            }
        )



class Signal:
    """A convinience class that encapsulates a signal in order to
       be used in the DisconnectedSignals context manager
    """

    def __init__(self, signal, receiver, sender):
        self.signal = signal
        self.receiver = receiver
        self.sender = sender

    def connect(self):
        self.signal.connect(receiver=self.receiver, sender=self.sender)

    def disconnect(self):
        self.signal.disconnect(receiver=self.receiver, sender=self.sender)


class DisconnectSignals:
    """A Context-Manager to temporarily disable a set of signals
    """

    def __init__(self, signals):
        self.signals = signals

    def __enter__(self):
        for signal in self.signals:
            signal.disconnect()

    def __exit__(self, type, value, traceback):
        for signal in self.signals:
            signal.connect()



from typing import Dict

from bridger.metadata.mixins import BridgerViewSetConfig


class PaginationBridgerViewSetConfig(BridgerViewSetConfig):

    def get_metadata(self) -> str:
        if hasattr(self.view, "pagination_class"):
            pagination = self.view.pagination_class.__name__ if self.view.pagination_class else None
            return {
                "CursorPagination": "cursor",
                "LimitOffsetPagination": "limitoffset",
                None: None
            }[pagination]
        return None

    @classmethod
    def get_metadata_key(cls):
        return "pagination"



import requests
from django.conf import settings
import pandas as pd

def fetch_df(endpoint):
    url = f"{settings.STAINLY_API_BASE_ENDPOINT_URL}{endpoint}"
    headers = {
        'user-agent': 'workbench',
        'Content-Type': 'application/json',
        'Accept': 'application/json'
    }
    if settings.STAINLY_TOKEN:
        headers["Authorization"] = f'Token {settings.STAINLY_TOKEN}'
    try:
        response = requests.get(url, headers=headers)
        if response.status_code == requests.codes.ok:
            r_json = response.json()
            if r_json.get("results", None):
                data = pd.json_normalize(r_json['results'])
                return data
    except requests.exceptions.ConnectionError as e:
        pass
    return pd.DataFrame()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bridger',
 'bridger.buttons',
 'bridger.clubhouse',
 'bridger.display',
 'bridger.docs',
 'bridger.endpoints',
 'bridger.filters',
 'bridger.fsm',
 'bridger.history',
 'bridger.markdown',
 'bridger.menus',
 'bridger.metadata',
 'bridger.metadata.configs',
 'bridger.migrations',
 'bridger.notifications',
 'bridger.notifications.viewsets',
 'bridger.pandas',
 'bridger.preview',
 'bridger.search',
 'bridger.serializers',
 'bridger.serializers.fields',
 'bridger.share',
 'bridger.signals',
 'bridger.tags',
 'bridger.tests',
 'bridger.tests.selenium',
 'bridger.titles',
 'bridger.utils',
 'bridger.viewsets',
 'bridger.websockets']

package_data = \
{'': ['*'], 'bridger': ['templates/bridger/*']}

install_requires = \
['celery[redis]>=4.4.0,<4.5.0',
 'channels>=2.4.0,<2.5.0',
 'channels_redis[cryptography]>=3.0.0,<3.1.0',
 'django-celery-beat>=2.0.0,<2.1.0',
 'django-filter>=2.2.0,<2.3.0',
 'django-fsm>=2.7.0,<2.8.0',
 'django-simple-history>=2.11.0,<2.12.0',
 'django>=3.0.0,<3.1.0',
 'djangorestframework-simplejwt>=4.3.0,<4.4.0',
 'djangorestframework>=3.11.0,<3.12.0',
 'inscriptis>=1.1,<2.0',
 'jsonschema>=3.2.0,<3.3.0',
 'markdown-blockdiag>=0.7.0,<0.8.0',
 'markdown>=3.2.0,<3.3.0',
 'numpy>=1.19.0,<1.20.0',
 'pandas>=1.0.0,<1.1.0',
 'pillow>=7.2.0,<7.3.0',
 'plotly>=4.8.0,<4.9.0',
 'psycopg2-binary>=2.8.0,<2.9.0',
 'pytest-selenium>=1.17.0,<1.18.0',
 'python-decouple==3.3',
 'python-slugify',
 'requests>=2.24.0,<2.25.0',
 'selenium>=3.141.0,<3.142.0',
 'termcolor>=1.1.0,<2.0.0',
 'toml>=0.10.0,<0.11.0']

setup_kwargs = {
    'name': 'django-bridger',
    'version': '0.10.11',
    'description': 'The bridge between a Django Backend and a Javascript based Frontend',
    'long_description': None,
    'author': 'Christopher Wittlinger',
    'author_email': 'c.wittlinger@intellineers.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.0,<3.9.0',
}


setup(**setup_kwargs)



from typing import Dict

from bridger.metadata.mixins import BridgerViewSetConfig
from bridger.serializers import ListSerializer, RepresentationSerializer, percent_decorator

class FieldsBridgerViewSetConfig(BridgerViewSetConfig):

    def get_metadata(self) -> str:
        if not hasattr(self.view, "get_serializer"):
            return None
            
        if serializer := self.view.get_serializer():
            fields = dict()
            rs = RepresentationSerializer
            ls = ListSerializer
            field_items = serializer.fields.items()

            for name, field in filter(lambda f: not isinstance(f[1], (rs, ls)), field_items):
                fields[name] = field.get_representation(self.request, name)

            for name, field in filter(lambda f: isinstance(f[1], (rs, ls)), field_items):
                representation = field.get_representation(self.request, name)
                fields[representation["related_key"]].update(representation)
                del fields[representation["related_key"]]["related_key"]

            serializer_class = self.view.get_serializer_class()

            for key, value in serializer_class.get_decorators():
                fields[key]["decorators"].append(value)
            for key in serializer_class.get_percent_fields():
                fields[key]["type"] = "percent"
                fields[key]["decorators"].append(percent_decorator)

            return fields

    @classmethod
    def get_metadata_key(cls):
        return "fields"



from typing import Optional

from rest_framework.request import Request

from bridger.display.instance_display import InstanceDisplay
from bridger.display.list_display import ListDisplay
from bridger.metadata.mixins import BridgerViewSetConfig


class DisplayConfig(BridgerViewSetConfig):
    def get_instance_display(self) -> Optional[InstanceDisplay]:
        return None

    def get_list_display(self) -> Optional[ListDisplay]:
        return None

    def get_preview_display(self) -> Optional[InstanceDisplay]:
        return None

    def get_metadata(self):
        yield "instance", list(self.get_instance_display() or [])
        yield "preview", list(self.get_preview_display() or [])
        if not self.instance:
            yield "list", dict(self.get_list_display() or {})

    @classmethod
    def get_metadata_key(cls):
        return "display"

    @classmethod
    def get_metadata_fieldname(cls):
        return "display_config_class"


class DisplayConfigMixin:
    display_config_class = DisplayConfig



from django.urls import path
from django.views.generic import TemplateView

from bridger.settings import bridger_settings


class FrontendView(TemplateView):
    template_name = bridger_settings.FRONTEND_TEMPLATE

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["BRIDGER_CONTEXT"] = bridger_settings.FRONTEND_CONTEXT
        return context

    @classmethod
    def bundled_view(cls, url_path):
        return path(route=url_path, view=cls.as_view(), name="frontend")



from django.utils import timezone
from rest_framework import filters
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.reverse import reverse

from bridger import buttons as bt
from bridger import display as dp
from bridger import viewsets
from bridger.enums import RequestType, WBIcon
from bridger.filters import DjangoFilterBackend

from bridger.notifications.models import Notification
from bridger.notifications.serializers import NotificationModelSerializer

from .buttons import NotificationButtonConfig
from .display import NotificationDisplayConfig
from .endpoint import NotificationEndpointConfig
from .title import NotificationTitleConfig

class NotificationModelViewSet(viewsets.ModelViewSet):

    queryset = Notification.objects.all()
    serializer_class = NotificationModelSerializer
    filter_backends = (
        filters.OrderingFilter,
        DjangoFilterBackend,
        filters.SearchFilter,
    )

    ordering_fields = ("timestamp_created",)
    ordering = ("-timestamp_created",)
    search_fields = ("title", "message")

    title_config_class = NotificationTitleConfig
    display_config_class = NotificationDisplayConfig
    endpoint_config_class = NotificationEndpointConfig
    button_config_class = NotificationButtonConfig

    def retrieve(self, request, *args, **kwargs):
        obj = self.get_object()
        if not obj.timestamp_read:
            obj.timestamp_read = timezone.now()
            obj.save()
        return super().retrieve(request, *args, **kwargs)

    def list(self, request, *args, **kwargs):
        self.get_queryset().filter(timestamp_received__isnull=True).update(timestamp_received=timezone.now())
        return super().list(request, *args, **kwargs)

    def get_queryset(self):
        return super().get_queryset().filter(recipient=self.request.user)

    @action(methods=["POST"], detail=False)
    def mark_all_as_read(self, request, pk=None):
        Notification.objects.filter(timestamp_read__isnull=True).update(timestamp_read=timezone.now())
        return Response({})

    @action(methods=["POST"], detail=False)
    def delete_all_read(self, request, pk=None):
        Notification.objects.filter(timestamp_read__isnull=False).delete()
        return Response({})



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from rest_framework.request import Request

from bridger.display import InstanceDisplay
from bridger.enums import RequestType
from bridger.serializers import ListSerializer, RepresentationSerializer, Serializer

from .bases import ButtonConfig, ButtonTypeMixin, ButtonUrlMixin
from .enums import ButtonType, HyperlinkTarget


@dataclass(unsafe_hash=True)
class DropDownButton(ButtonTypeMixin, ButtonConfig):
    button_type = ButtonType.DROPDOWN
    buttons: Tuple = field(default_factory=tuple)

    def __post_init__(self):
        if hasattr(super(), "__post_init__"):
            super().__post_init__()
        assert isinstance(self.buttons, tuple)

    def get_buttons(self):
        # All buttons nested inside this dropdown button need to be handled by converting it to a dictionairy
        for button in self.buttons:
            yield dict(button)

    def __iter__(self):
        yield from super().__iter__()
        yield "buttons", self.get_buttons()


@dataclass(unsafe_hash=True)
class WidgetButton(ButtonTypeMixin, ButtonUrlMixin, ButtonConfig):
    button_type = ButtonType.WIDGET
    new_mode: bool = False

    def __iter__(self):
        yield from super().__iter__()
        yield "new_mode", self.new_mode


@dataclass(unsafe_hash=True)
class HyperlinkButton(ButtonTypeMixin, ButtonUrlMixin, ButtonConfig):
    button_type = ButtonType.HYPERLINK
    target: HyperlinkTarget = HyperlinkTarget.BLANK

    def __iter__(self):
        yield from super().__iter__()
        yield "target", self.target.value


@dataclass(unsafe_hash=True)
class ActionButton(ButtonTypeMixin, ButtonUrlMixin, ButtonConfig):
    button_type = ButtonType.ACTION
    method: RequestType = RequestType.POST
    action_label: str = ""

    description_fields: str = "<p>Are you sure you want to proceed?</p>"
    instance_display: InstanceDisplay = None
    serializer: Serializer = None
    confirm_config: ButtonConfig = ButtonConfig(label="Confirm", title="Confirm")
    cancel_config: ButtonConfig = ButtonConfig(label="Cancel", title="Cancel")

    identifiers: Tuple[str] = field(default_factory=tuple)

    def __post_init__(self):
        if hasattr(super(), "__post_init__"):
            super().__post_init__()
        assert isinstance(self.identifiers, tuple)

    def get_fields(self, request: Request) -> Dict:
        fields = dict()
        rs = RepresentationSerializer
        ls = ListSerializer
        field_items = self.serializer().fields.items()

        for name, field in filter(lambda f: not isinstance(f[1], (rs, ls)), field_items):
            fields[name] = field.get_representation(request, name)

        for name, field in filter(lambda f: isinstance(f[1], (rs, ls)), field_items):
            representation = field.get_representation(request, name)
            fields[representation["related_key"]].update(representation)
            del fields[representation["related_key"]]["related_key"]

        return fields

    def _get_fields(self, request: Request) -> Dict:
        fields = self.get_fields(request)
        serializer_class = self.serializer
        return fields

    def __iter__(self):
        yield from super().__iter__()
        yield "action_label", self.action_label
        yield "method", self.method.value

        yield "description_fields", self.description_fields
        yield "confirm_config", dict(self.confirm_config)
        yield "cancel_config", dict(self.cancel_config)

        yield "identifiers", self.identifiers

        if self.instance_display:
            yield "instance_display", list(self.instance_display)

        if self.serializer:
            assert (
                hasattr(self, "request") and self.request is not None
            ), "Action Buttons who define a custom serializer, needs to have access to the request"
            yield "fields", self._get_fields(self.request)



from enum import Enum

class ChoiceEnum(Enum):

    @classmethod
    def choices(cls):
        return [(choice.name, choice.value) for choice in cls]

    @classmethod
    def names(cls):
        return [choice.name for choice in cls]
    
    @classmethod
    def values(cls):
        return [choice.value for choice in cls]

    @classmethod
    def to_dict(cls):
        return {choice.name: choice.value for choice in cls}





from asgiref.sync import async_to_sync
from celery import shared_task
from channels.layers import get_channel_layer
from django.conf import settings
from django.core import mail
from django.core.mail import EmailMultiAlternatives
from django.template.loader import get_template
from django.utils import timezone
from django.utils.html import strip_tags

from .models import Notification


@shared_task(name="bridger.notifications.send_system")
def send_system(notification_id):
    notification = Notification.objects.get(id=notification_id)
    channel_layer_name = f"notification-{notification.recipient.id}"
    async_to_sync(get_channel_layer().group_send)(
        channel_layer_name, {"notification_id": notification.id, "type": "notification.notify"},
    )


@shared_task(name="bridger.notifications.send_mail")
def send_mail(notification_id):
    notification = Notification.objects.get(id=notification_id)
    context = {"notification": notification}
    template = getattr(settings, "BRIDGER_NOTIFICATION_TEMPLATE", "bridger/email_notification_template.html",)
    email_to = notification.recipient.email
    email_from = getattr(settings, "BRIDGER_NOTIFICATION_EMAIL_FROM", "system@bridger.com",)
    rendered_template = get_template(template).render(context)

    msg = EmailMultiAlternatives(
        subject=f"Notification {notification.title}", body=strip_tags(rendered_template), from_email=email_from, to=[email_to],
    )
    msg.attach_alternative(rendered_template, "text/html")
    msg.send()

    Notification.objects.filter(id=notification.id).update(timestamp_mailed=timezone.now())
